
#projectname {

}

code {
  font-family: Courier New, monospace, fixed;
  font-size: 1em;
}

dt {
  font-weight: bold;
}

div.multicol {
  -moz-column-gap: 1em;
  -webkit-column-gap: 1em;
  -moz-column-count: 3;
  -webkit-column-count: 3;
}

p.startli, p.startdd, p.starttd {
  margin-top: 2px;
}

p.endli {
  margin-bottom: 0px;
}

p.enddd {
  margin-bottom: 4px;
}

p.endtd {
  margin-bottom: 2px;
}

/* @end */

caption {
  font-weight: bold;
}

span.legend {
        font-size: 70%;
        text-align: center;
}

h3.version {
        font-size: 90%;
        text-align: center;
}

div.qindex, div.navtab{
  background-color: #EBEFF6;
  border: 1px solid #A3B4D7;
  text-align: center;
}

div.qindex, div.navpath {
  width: 100%;
  line-height: 140%;
}

div.navtab {
  margin-right: 15px;
}

/* @group Link Styling */

a {
  color: #3D578C;
  font-weight: normal;
  text-decoration: none;
}

.contents a:visited {
  color: #4665A2;
}

a:hover {
  text-decoration: underline;
}

a.qindex {
  font-weight: bold;
}

a.qindexHL {
  font-weight: bold;
  background-color: #9CAFD4;
  color: #ffffff;
  border: 1px double #869DCA;
}

.contents a.qindexHL:visited {
        color: #ffffff;
}

a.el {
  font-weight: bold;
}

a.elRef {
}

a.code, a.code:visited {
  color: #4665A2; 
}

a.codeRef, a.codeRef:visited {
  color: #4665A2; 
}

/* @end */

dl.el {
  margin-left: -1cm;
}

a.code {
  font-family: Courier New, monospace, fixed;
  font-size: 1em;
}


span.lineno {
  padding-right: 4px;
  text-align: right;
  border-right: 2px solid #0F0;
  background-color: #E8E8E8;
        white-space: pre;
}
span.lineno a {
  background-color: #D8D8D8;
}

span.lineno a:hover {
  background-color: #C8C8C8;
}

div.ah {
  background-color: black;
  font-weight: bold;
  color: #ffffff;
  margin-bottom: 3px;
  margin-top: 3px;
  padding: 0.2em;
  border: solid thin #333;
  border-radius: 0.5em;
  -webkit-border-radius: .5em;
  -moz-border-radius: .5em;
  box-shadow: 2px 2px 3px #999;
  -webkit-box-shadow: 2px 2px 3px #999;
  -moz-box-shadow: rgba(0, 0, 0, 0.15) 2px 2px 2px;
  background-image: -webkit-gradient(linear, left top, left bottom, from(#eee), to(#000),color-stop(0.3, #444));
  background-image: -moz-linear-gradient(center top, #eee 0%, #444 40%, #000);
}

div.groupHeader {
  margin-left: 16px;
  margin-top: 12px;
  font-weight: bold;
}

div.groupText {
  margin-left: 16px;
  font-style: italic;
}

body {
  background-color: white;
  color: black;
        margin: 0;
}


td.indexkey {
  background-color: #EBEFF6;
  font-weight: bold;
  border: 1px solid #C4CFE5;
  margin: 2px 0px 2px 0;
  padding: 2px 10px;
        white-space: nowrap;
        vertical-align: top;
}

td.indexvalue {
  background-color: #EBEFF6;
  border: 1px solid #C4CFE5;
  padding: 2px 10px;
  margin: 2px 0px;
}

tr.memlist {
  background-color: #EEF1F7;
}

p.formulaDsp {
  text-align: center;
}

img.formulaDsp {
  
}

img.formulaInl {
  vertical-align: middle;
}

div.center {
  text-align: center;
        margin-top: 0px;
        margin-bottom: 0px;
        padding: 0px;
}

div.center img {
  border: 0px;
}

address.footer {
  text-align: right;
  padding-right: 12px;
}

img.footer {
  border: 0px;
  vertical-align: middle;
}

/* @group Code Colorization */

span.keyword {
  color: #008000
}

span.keywordtype {
  color: #604020
}

span.keywordflow {
  color: #e08000
}

span.comment {
  color: #800000
}

span.preprocessor {
  color: #806020
}

span.stringliteral {
  color: #002080
}

span.charliteral {
  color: #008080
}

span.vhdldigit { 
  color: #ff00ff 
}

span.vhdlchar { 
  color: #000000 
}

span.vhdlkeyword { 
  color: #700070 
}

span.vhdllogic { 
  color: #ff0000 
}

blockquote {
        background-color: #F7F8FB;
        border-left: 2px solid #9CAFD4;
        margin: 0 24px 0 4px;
        padding: 0 12px 0 16px;
}

/* @end */


td.tiny {
  font-size: 75%;
}

.dirtab {
  padding: 4px;
  border-collapse: collapse;
  border: 1px solid #A3B4D7;
}

th.dirtab {
  background: #EBEFF6;
  font-weight: bold;
}

hr {
  height: 0px;
  border: none;
  border-top: 1px solid #4A6AAA;
}

hr.footer {
  height: 1px;
}

/* @group Member Descriptions */

table.memberdecls {
  border-spacing: 0px;
  padding: 0px;
}

.memTemplParams {
  color: #4665A2;
        white-space: nowrap;
}

/* @end */

/* @group Member Details */

/* Styles for detailed member documentation */

.memtemplate {
  font-size: 80%;
  color: #4665A2;
  font-weight: normal;
  margin-left: 9px;
}

.memnav {
  background-color: #EBEFF6;
  border: 1px solid #A3B4D7;
  text-align: center;
  margin: 2px;
  margin-right: 15px;
  padding: 2px;
}

.mempage {
  width: 100%;
}


address {
  font-style: normal;
  color: #2A3D61;
}

.tabsearch {
  top: 0px;
  left: 10px;
  height: 36px;
  background-image: url('tab_b.png');
  z-index: 101;
  overflow: hidden;
  font-size: 13px;
}

div.summary
{
  float: right;
  font-size: 8pt;
  padding-right: 5px;
  width: 50%;
  text-align: right;
}       

div.summary a
{
  white-space: nowrap;
}

dl
{
  padding: 0 0 0 10px;
}




#projectlogo
{
  text-align: center;
  vertical-align: bottom;
  border-collapse: separate;
}
 
#projectlogo img
{ 
  border: 0px none;
}
 
#projectname
{
  font-size: 1.5em;
  font-weight: 600;
  margin: 0px;
  padding: 2px 0px;
}
    
#projectbrief
{
  font-size: 1.2em;
  margin: 0px;
  padding: 0px;
}

#projectnumber
{
  font-size: 1em;
  margin: 0px;
  padding: 0px;
}

#titlearea
{
  padding: 0px;
  margin: 0px;
  width: 100%;
  border-bottom: 1px solid #5373B4;
}

.image
{
        text-align: center;
}

.dotgraph
{
        text-align: center;
}

.mscgraph
{
        text-align: center;
}

.caption
{
  font-weight: bold;
}

div.zoom
{
  border: 1px solid #90A5CE;
}

dl.citelist {
        margin-bottom:50px;
}

dl.citelist dt {
        color:#334975;
        float:left;
        font-weight:bold;
        margin-right:10px;
        padding:5px;
}

dl.citelist dd {
        margin:2px 0;
        padding:5px 0;
}

div.toc {
        padding: 14px 25px;
        background-color: #F4F6FA;
        border: 1px solid #D8DFEE;
        border-radius: 7px 7px 7px 7px;
        float: right;
        height: auto;
        margin: 0 20px 10px 10px;
        width: 200px;
}

div.toc li {
        background: url("bdwn.png") no-repeat scroll 0 5px transparent;
        font: 10px/1.2 Verdana,DejaVu Sans,Geneva,sans-serif;
        margin-top: 5px;
        padding-left: 10px;
        padding-top: 2px;
}

div.toc h3 {
        font: bold 12px/1.2 Arial,FreeSans,sans-serif;
  color: #4665A2;
        border-bottom: 0 none;
        margin: 0;
}

div.toc ul {
        list-style: none outside none;
        border: medium none;
        padding: 0px;
}       

div.toc li.level1 {
        margin-left: 0px;
}

div.toc li.level2 {
        margin-left: 15px;
}

div.toc li.level3 {
        margin-left: 30px;
}

div.toc li.level4 {
        margin-left: 45px;
}

.inherit_header {
  font-weight: bold;
  color: gray;
  cursor: pointer;
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -khtml-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}

.inherit_header td {
        padding: 6px 0px 2px 5px;
}

.inherit {
        display: none;
}

tr.heading h2 {
        margin-top: 12px;
        margin-bottom: 4px;
}

@media print
{
  #top { display: none; }
  #side-nav { display: none; }
  #nav-path { display: none; }
  body { overflow:visible; }
  h1, h2, h3, h4, h5, h6 { page-break-after: avoid; }
  .summary { display: none; }
  .memitem { page-break-inside: avoid; }
  #doc-content
  {
    margin-left:0 !important;
    height:auto !important;
    width:auto !important;
    overflow:inherit;
    display:inline;
  }
}
