EXPORTS
  lua_tolstring
  lua_typename
  lua_pushfstring
  lua_pushvfstring
  lua_getlocal
  lua_getupvalue
  lua_setlocal
  lua_setupvalue
  lua_topointer
  lua_iscfunction
  lua_isnumber
  lua_isstring
  lua_isuserdata
  lua_toboolean
  lua_type
  lua_equal
  lua_lessthan
  lua_rawequal
  lua_checkstack
  lua_cpcall
  lua_error
  lua_getmetatable
  lua_gettop
  lua_load
  lua_next
  lua_pcall
  lua_pushthread
  lua_setfenv
  lua_setmetatable
  lua_resume
  lua_status
  lua_yield
  lua_dump
  lua_gc
  lua_gethook
  lua_gethookcount
  lua_gethookmask
  lua_getinfo
  lua_getstack
  lua_sethook
  lua_getallocf
  lua_tocfunction
  lua_atpanic
  lua_tointeger
  lua_tonumber
  lua_tothread
  lua_newstate
  lua_newthread
  lua_objlen
  lua_touserdata
  lua_close
  lua_call
  lua_concat
  lua_createtable
  lua_getfenv
  lua_getfield
  lua_gettable
  lua_insert
  lua_pushboolean
  lua_pushcclosure
  lua_pushinteger
  lua_pushlightuserdata
  lua_pushlstring
  lua_pushnil
  lua_pushnumber
  lua_pushstring
  lua_pushvalue
  lua_rawget
  lua_rawgeti
  lua_rawset
  lua_rawseti
  lua_remove
  lua_replace
  lua_setfield
  lua_settable
  lua_settop
  lua_xmove
  lua_newuserdata
  lua_setallocf
  luaL_prepbuffer
  luaL_checklstring
  luaL_findtable
  luaL_gsub
  luaL_optlstring
  luaL_newmetatable
  luaL_argerror
  luaL_callmeta
  luaL_checkoption
  luaL_error
  luaL_getmetafield
  luaL_loadbuffer
  luaL_loadfile
  luaL_loadstring
  luaL_ref
  luaL_typerror
  luaL_checkinteger
  luaL_optinteger
  luaL_checknumber
  luaL_optnumber
  luaL_newstate
  luaL_openlib
  luaL_addlstring
  luaL_addstring
  luaL_addvalue
  luaL_buffinit
  luaL_checkany
  luaL_checkstack
  luaL_checktype
  luaL_pushresult
  luaL_register
  luaL_unref
  luaL_where
  luaL_checkudata
  luaopen_base
  luaopen_debug
  luaopen_io
  luaopen_math
  luaopen_os
  luaopen_package
  luaopen_string
  luaopen_table
  luaL_openlibs
  luaU_dump
  luaM_toobig
  luaM_realloc_
  luaS_newlstr
  luaD_growstack
  luaF_newproto
  luaP_opmodes
  luaP_opnames
