0  ICON  "..\\etc\\luajit.ico"

1 VERSIONINFO
 FILEVERSION 2,0,0,0
 PRODUCTVERSION 2,0,0,0
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "www.luajit.org\0"
            VALUE "CompanyName", "luajit.org\0"
            VALUE "FileDescription", "LuaJIT is a JIT copmiler for the Lua language\0"
            VALUE "FileVersion", "2.0.0-beta10\0"
            VALUE "LegalCopyright", "Copyright © Mike Pall.\0"
            VALUE "OriginalFilename", "luajit.exe\0"
            VALUE "ProductName", "LuaJIT\0"
            VALUE "ProductVersion", "2.0.0-beta10\0"
            VALUE "PrivateBuild", "Built for LuaDist\0"
        END
    END
END

#ifdef MSVC8
1 24 "lua_dll8.manifest"
#elif MSVC9
1 24 "lua_dll9.manifest"
#endif
