/* 
 * Copyright 2008-2018 Aerospike, Inc.
 *
 * Portions may be licensed to Aerospike, Inc. under one or more contributor
 * license agreements.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
#pragma once

#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <stdio.h>



/**
 * Registers an Object
 * An Object is a Lua Table that is not bound to userdata.
 */
int mod_lua_reg_object(lua_State *, const char *, const luaL_reg *, const luaL_reg *);

/**
 * Registers a Class
 * A Class is a Lua Table that is bound to userdata.
 */
int mod_lua_reg_class(lua_State *, const char *, const luaL_reg *, const luaL_reg *);
