
body {
  font-family: Helvetica, Arial, Sans-Serif;
  font-size: 14px;
  font-style: normal;
  font-weight: 400;
}

.footer {
  border-top: 1px solid #ddd;
  color: #8F8F8F;
  font-size: 0.9em;
  padding: 16px;
  text-align: center;
}

.footer > .copy {
  font-weight: 700;
}

.footer > .legal {
}

.footer > .legal > a {
  color: inherit;
}

.footer > .legal > span {
}

.footer > .gen {
  color: #BBB;
  margin-top: 24px;
}

/*******************************************************************************
 * Top
 ******************************************************************************/

#top {
}

#top > #titlearea {
  margin-top: 16px;
  border: none;
}

#projectlogo {
}

#projectlogo > img {
  height: 48px;
}

#projectname {
  border-left: 1px solid #FF6666;
  color: #fff;
  font-size: 1.5em;
  font-weight: normal;
  padding-left: 24px;
  text-shadow: 1px 1px 1px #4E090A;
}



/*******************************************************************************
 * Headings
 ******************************************************************************/

h1, h2, h3, h4, h5, h6 {
  border: none;
  color: #444;
  font-family: inherit;
  font-size: inherit;
  font-style: normal;
  font-weight: 400;
  margin: 1em 0 1em 0;
  padding: 0;
}

h1 {
  font-size: 1.6em;
}

h2 {
  font-size: 1.2em;
}

h2.groupheader {
  border-radius: 4px;
  color: #eee;

  background: #4f4f4f; /* Old browsers */
  background: -moz-linear-gradient(top,  #4f4f4f 0%, #212121 100%); /* FF3.6+ */
  background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#4f4f4f), color-stop(100%,#212121)); /* Chrome,Safari4+ */
  background: -webkit-linear-gradient(top,  #4f4f4f 0%,#212121 100%); /* Chrome10+,Safari5.1+ */
  background: -o-linear-gradient(top,  #4f4f4f 0%,#212121 100%); /* Opera 11.10+ */
  background: -ms-linear-gradient(top,  #4f4f4f 0%,#212121 100%); /* IE10+ */
  background: linear-gradient(to bottom,  #4f4f4f 0%,#212121 100%); /* W3C */
  filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#4f4f4f', endColorstr='#212121',GradientType=0 ); /* IE6-9 */

  padding: 8px;
  text-shadow: -1px -1px 1px #000;
}

/*******************************************************************************
 * Header
 ******************************************************************************/

.header {
  /*background: #eee;*/
  /*border-bottom: 1px solid #ddd;*/
  margin: 16px;
}

.header > .summary {
  padding: 16px 0;
  font-size: inherit;
}

.header > .summary > a {
  font-size: inherit;
  font-weight: 400;
  white-space: nowrap;
}

.header > .headertitle {

}

.header > .headertitle > .title {
  font-size: 2em;
}

.header > .headertitle > .title > .ingroups {
  font-size: 14px;
}

.header > .headertitle > .title > .ingroups > a {
  margin-right: 16px;
  font-size: inherit;
  font-weight: 400;
  white-space: nowrap;
}

/*******************************************************************************
 * Content
 ******************************************************************************/

.contents {
  margin: 16px;
  padding: 0px;
}

.contents > h2.groupheader {
  font-size: 1.2em;
  font-weight: normal;
  margin-top: 0px !important;
}

.contents > p {
  margin: 1em 0;
}

.contents > h1, .contents > h2, .contents > h3, .contents > h4 {
  margin: 1em 0;
}

.contents > h1 {
  font-size: 1.6em;
  font-weight: bold;
}

.contents > h2 {
  font-size: 1.4em;
  font-weight: bold;
}

.contents > h3 {
  font-size: 1.2em;
  font-weight: bold;
}

./*contents > .fragment {
  margin: 1em 0;
}*/

/*.contents > .doxtable {
  margin: 1em 0;
}
*/

.contents > .textblock {
  border-bottom: 1px solid #ddd;
}


.contents > .textblock > h1, .contents > h2, .contents > h3, .contents > h4 {
  margin: 1em 0;
}

.contents > .textblock > h1 {
}

.contents > .textblock > h2 {
  font-size: 1.4em;
  font-weight: bold;
}

.contents > .textblock > h3 {
  font-size: 1.2em;
  font-weight: bold;
}

.dynheader {
  border-top: 1px solid #ddd;
  padding: 16px 0;
}

/*******************************************************************************
 * Fragment
 ******************************************************************************/

.fragment {
  font-family: "Consolas", Courier New, monospace, fixed;
  background: #f5f5f5;
  border: 1px solid #dddddd;
  border-radius: 4px;
  font-size: 1em;
  padding: 16px;
}

.fragment > .line {
  background: none;
  font-family: inherit;
  font-size: inherit;
  line-height: 1.2em;
  /*padding: 8px;*/
  min-height: 1.2em;
  text-wrap: unrestricted;
  white-space: -moz-pre-wrap; /* Moz */
  white-space: -pre-wrap;     /* Opera 4-6 */
  white-space: -o-pre-wrap;   /* Opera 7 */
  white-space: pre-wrap;      /* CSS3  */
  word-wrap: break-word;      /* IE 5.5+ */
}

/*******************************************************************************
 * Member Decls
 ******************************************************************************/

table.memberdecls {
  margin: 16px 0;
}

table.memberdecls > tbody > tr > td {
  border-bottom: 1px solid #dddddd;
  padding: 8px;
}

table.memberdecls > tbody > tr.heading > td {
  border: none;
  padding: 0;
}

table.memberdecls > tbody > tr.heading > td * {
}

table.memberdecls > tbody > tr > td.ititle {
  background: #eee;
}

table.memberdecls > tbody > tr > td.ititle p:first-child{
  margin: 0;
}

table.memberdecls > tbody > tr > td.memSeparator {
  display: none;
}

table.memberdecls > tbody > tr > td.memItemLeft {
  white-space: nowrap;
  width: auto;
}

table.memberdecls > tbody > tr > td.memItemRight {
  width: 100%;
}

/*******************************************************************************
 * Member Item
 ******************************************************************************/

div.memitem {
  border: 1px solid #dddddd;
  border-radius: 4px;
  margin: 16px 0;
}

div.memitem > div.memproto {
  color: #444;
  text-shadow: 1px 1px 1px #e5e5e5;

  background: #dfdfdf; /* Old browsers */
  background: -moz-linear-gradient(top,  #dfdfdf 0%, #c3c3c3 100%); /* FF3.6+ */
  background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#dfdfdf), color-stop(100%,#c3c3c3)); /* Chrome,Safari4+ */
  background: -webkit-linear-gradient(top,  #dfdfdf 0%,#c3c3c3 100%); /* Chrome10+,Safari5.1+ */
  background: -o-linear-gradient(top,  #dfdfdf 0%,#c3c3c3 100%); /* Opera 11.10+ */
  background: -ms-linear-gradient(top,  #dfdfdf 0%,#c3c3c3 100%); /* IE10+ */
  background: linear-gradient(to bottom,  #dfdfdf 0%,#c3c3c3 100%); /* W3C */
  filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#dfdfdf', endColorstr='#c3c3c3',GradientType=0 ); /* IE6-9 */

  border-bottom: 1px solid #aaa;
  padding: 8px;
}



div.memitem > div.memproto > table.mlabels {
  border-spacing: 0px;
}

div.memitem > div.memproto > table.mlabels td.mlabels-left {
  width: 100%;
  padding: 0px;
}

div.memitem > div.memproto > table.mlabels td.mlabels-right {
  vertical-align: bottom;
  padding: 0px;
  white-space: nowrap;
}

div.memitem > div.memproto > table.mlabels td.mlabels-right span.mlabels {
  margin-left: 8px;
}

span.mlabels > span.mlabel {
  color: white;
  background-color: #728DC1;
  border-top:1px solid #5373B4;
  border-left:1px solid #5373B4;
  border-right:1px solid #C4CFE5;
  border-bottom:1px solid #C4CFE5;
  border-radius: 4px;
  font-size: 0.8em;
  margin-left: 4px;
  padding: 3px 8px;
  text-shadow: 1px 1px 1px #516BA8;
  white-space: nowrap;
}



div.memitem > div.memdoc {
  padding: 16px;
}

div.memitem > div.memdoc > *:first-child {
  margin-top: 0;
}

div.memitem > div.memdoc > *:last-child {
  margin-bottom: 0;
}

.memname {
  font-weight: bold;
  margin-left: 6px;
}

.memname td {
  vertical-align: bottom;
}


.paramkey {
  text-align: right;
}

.paramtype {
  white-space: nowrap;
}

.paramname {
  white-space: nowrap;
}
.paramname em {
  font-style: normal;
}

.paramname code {
  line-height: 14px;
}

.params, .retval, .exception, .tparams {
  margin-left: 0px;
  padding-left: 0px;
}       

.params .paramname, .retval .paramname {
  /*font-weight: bold;*/
  vertical-align: top;
}

.params .paramtype {
  font-style: italic;
  vertical-align: top;
}       

.params .paramdir {
  vertical-align: top;
}

a.el {
  font-weight: bold;
}

/*******************************************************************************
 * Field Table
 ******************************************************************************/

table.fieldtable {
  border: 1px solid #EEE;
}

table.fieldtable > tbody > tr > th {
  background: #e5e5e5;
  padding: 8px;
  text-align: left;
}

table.fieldtable > tbody > tr > td {
  border: 1px solid #EEE;
  padding: 8px;
  vertical-align: top !important;
}

table.fieldtable > tbody > tr > td.fieldname {
  font-family: Courier New, monospace;
  font-style: normal !important;
}

table.fieldtable > tbody > tr > td.fieldname * {
  font-style: normal !important;
}

table.fieldtable > tbody > tr > td.fielddoc {
  padding: 8px;
}

table.fieldtable > tbody > tr > td.fielddoc p:first-child {
  margin: 0;
}

/*******************************************************************************
 * Doxtable
 ******************************************************************************/

.doxtable {
  border: 1px solid #ddd;
}

.doxtable > tbody > tr > th {
  background: #e5e5e5;
  font-weight: 400;
  padding: 8px;
}

.doxtable > tbody > tr > td {
  border: 1px solid #eee;
  font-weight: 400;
  padding: 8px;
}


/*******************************************************************************
 * Navbar
 ******************************************************************************/

div.tabs, div.tabs2, div.tabs3 {
  width: 100%;
  z-index: 101;
  background: #3d3d3d; /* Old browsers */
  background: -moz-linear-gradient(top,  #3d3d3d 0%, #0e0e0e 100%); /* FF3.6+ */
  background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#3d3d3d), color-stop(100%,#0e0e0e)); /* Chrome,Safari4+ */
  background: -webkit-linear-gradient(top,  #3d3d3d 0%,#0e0e0e 100%); /* Chrome10+,Safari5.1+ */
  background: -o-linear-gradient(top,  #3d3d3d 0%,#0e0e0e 100%); /* Opera 11.10+ */
  background: -ms-linear-gradient(top,  #3d3d3d 0%,#0e0e0e 100%); /* IE10+ */
  background: linear-gradient(to bottom,  #3d3d3d 0%,#0e0e0e 100%); /* W3C */
  filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#3d3d3d', endColorstr='#0e0e0e',GradientType=0 ); /* IE6-9 */
}

ul.tablist {
  border: 0;
  margin: 0;
  padding: 0;
  display: block;
}

ul.tablist > li {
  margin: 0;
  padding: 0;
  display: table-cell;
  line-height: 2.5em;
  border-right: 1px solid rgba(0,0,0,0.4);
  border-left: 1px solid rgba(255,255,255,0.1);
}

ul.tablist > li > a {
  color: #fff;
  display: block;
  margin: 0;
  outline: none;
  padding: 0 20px;
  text-shadow: -1px -1px 1px #000;
}

ul.tablist > li > a:hover {
  background: rgba(255,255,255,0.1);
}

ul.tablist > li.current {
  background: rgba(255,255,255,0.2);
}


/*******************************************************************************
 * Directory
 ******************************************************************************/

div.directory {
  margin: 10px 0px;
  /*border-top: 1px solid #A8B8D9;*/
  /*border-bottom: 1px solid #A8B8D9;*/
  width: 100%;
}

div.directory > div.levels {
  white-space: nowrap;
  width: 100%;
  text-align: right;
  font-size: 9pt;
}

div.directory > div.levels span {
  cursor: pointer;
  padding-left: 2px;
  padding-right: 2px;
  color: #3D578C;
}

div.directory > table.directory {
  border-collapse:collapse;
}

table.directory > tbody > tr > td {
  margin: 0px;
  padding: 4px;
  vertical-align: top;
}

table.directory > tbody > tr > td.entry {
  white-space: nowrap;
  padding-right: 6px;
}

table.directory > tbody > tr > td.entry a {
  outline:none;
}

table.directory > tbody > tr > td.entry a img {
  border: none;
}

table.directory > tbody > tr > td.desc {
  width: 100%;
  padding-left: 6px;
  padding-right: 6px;
  padding-top: 3px;
  /*border-left: 1px solid rgba(0,0,0,0.05);*/
}

table.directory > tbody > tr.even {
  background-color: #f5f5f5;
}

.directory img {
  vertical-align: -30%;
}

/*******************************************************************************
 * Sections / Messages
 ******************************************************************************/

dl.section {
  margin: 8px 0;
  padding: 16px;
}

dl.section dd {
  /*margin-bottom: 8px;*/
}

dl.note {
  background: #FAF9E6;
  border-left: 8px solid #D0C000;
  padding: 8px 16px;
}

dl.warning, dl.attention {
  background: #FFE6E6;
  border-left: 4px solid #FF0000;
  padding: 8px 16px;
}

dl.pre, dl.post, dl.invariant {
  border-left: 4px solid #00D000;
}

dl.deprecated {
  background: #FFE6E6;
  border-left: 4px solid #505050;
  padding: 8px 16px;
}

dl.todo {
  border-left: 4px solid #00C0E0;
}

dl.test {
  border-left: 4px solid #3030E0;
}

dl.bug {
  border-left: 4px solid #C08050;
}


dl.section.return {
  margin: 16px 0;
  padding: 0;
}


/*******************************************************************************
 * Nav Path
 ******************************************************************************/

#nav-path {
  display: none;
}

.navpath ul {
  background: #0e0e0e;
  height:30px;
  line-height:30px;
  overflow:hidden;
  margin:0px;
  padding:0px;
}

.navpath li {
  list-style-type:none;
  float:left;
  padding-left:10px;
  padding-right:15px;
  background: #fff;
}

.navpath li.navelem a {
  height: 32px;
  display:block;
  text-decoration: none;
  outline: none;
}

.navpath li.navelem a:hover {
  color:#6884BD;
}

.navpath li.footer {
  list-style-type:none;
  float:right;
  padding-left:10px;
  padding-right:15px;
  background-image:none/*;
  background-repeat:no-repeat;
  background-position:r*/ight;
  color:#364D7C;
  font-size: 8pt;
}

/*******************************************************************************
 * Reflist
 ******************************************************************************/

dl.reflist {
  /*background: #CCC;*/
  border-bottom: 1px solid #ddd;
  padding: 0;
}

dl.reflist > dt {
  border-top: 1px solid #ddd;
  padding: 8px;
}

dl.reflist > dd {
  padding: 8px;
}

