#define CK_F_PR_CAS_64
#define CK_F_PR_CAS_64_VALUE
#define CK_F_PR_CAS_PTR
#define CK_F_PR_CAS_PTR_VALUE
#define CK_F_PR_FAS_32
#define CK_F_PR_FAS_UINT
#define CK_F_PR_FAS_INT
#define CK_F_PR_CAS_32
#define CK_F_PR_CAS_32_VALUE
#define CK_F_PR_CAS_UINT
#define CK_F_PR_CAS_INT
#define CK_F_PR_CAS_UINT_VALUE
#define CK_F_PR_CAS_INT_VALUE
#define CK_F_PR_STORE_64
#define CK_F_PR_STORE_32
#define CK_F_PR_STORE_DOUBLE
#define CK_F_PR_STORE_UINT
#define CK_F_PR_STORE_INT
#define CK_F_PR_STORE_PTR
#define CK_F_PR_LOAD_64
#define CK_F_PR_LOAD_32
#define CK_F_PR_LOAD_DOUBLE
#define CK_F_PR_LOAD_UINT
#define CK_F_PR_LOAD_INT
#define CK_F_PR_LOAD_PTR

