1 VERSIONINFO
 FILEVERSION 5,1,4,0
 PRODUCTVERSION 5,1,4,0
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", "www.lua.org\0"
            VALUE "CompanyName", "Lua.org\0"
            VALUE "FileDescription", "Lua Language Run Time\0"
            VALUE "FileVersion", "5.1.4\0"
            VALUE "LegalCopyright", "Copyright © 1994-2009 Lua.org, PUC-Rio.\0"
            VALUE "OriginalFilename", "lua5.1.dll\0"
            VALUE "ProductName", "Lua - The Programming Language\0"
            VALUE "ProductVersion", "5.1.4\0"
            VALUE "PrivateBuild", "Built for LuaDist\0"
        END
    END
END

#ifdef MSVC8
2 24 "lua_dll8.manifest"
#elif MSVC9
2 24 "lua_dll9.manifest"
#endif
